import os
from typing import Any, ClassVar, Optional

from dotenv import load_dotenv
from neo4j import GraphDatabase
from neo4j import Result as Neo4jResult
from pydantic import model_validator

from ..result import NeontologyResult
from .graphengine import GraphEngineBase, GraphEngineConfig
from .neo4jengine import neo4j_records_to_neontology_records


class MemgraphEngine(GraphEngineBase):
    """
    Note. the memgraph engine is very similar to the Neo4j engine as they
        can both use the same Neo4j driver
    Args:
        Neo4jEngine (_type_): _description_
    """

    def __init__(self, config: "MemgraphConfig") -> None:
        """Initialise connection to the engine

        Args:
            config - Takes a MemgraphConfig object.
        """

        self.driver = GraphDatabase.driver(  # type: ignore
            config.uri,
            auth=(config.username, config.password),
        )

    def verify_connection(self) -> bool:
        try:
            self.driver.verify_connectivity()
            return True
        except:  # noqa: E722
            return False

    def close_connection(self) -> None:
        try:
            self.driver.close()
        except AttributeError:
            pass

    def evaluate_query(
        self,
        cypher: str,
        params: dict = {},
        node_classes: dict = {},
        relationship_classes: dict = {},
    ) -> NeontologyResult:
        result = self.driver.execute_query(cypher, parameters_=params)

        neo4j_records = result.records
        neontology_records, nodes, rels, paths = neo4j_records_to_neontology_records(
            neo4j_records, node_classes, relationship_classes
        )

        return NeontologyResult(
            records_raw=neo4j_records,
            records=neontology_records,
            nodes=nodes,
            relationships=rels,
            paths=paths,
        )

    def evaluate_query_single(self, cypher: str, params: dict = {}) -> Optional[Any]:
        result = self.driver.execute_query(
            cypher, parameters_=params, result_transformer_=Neo4jResult.single
        )

        if result:
            return result.value()

        else:
            return None


class MemgraphConfig(GraphEngineConfig):
    engine: ClassVar[type[MemgraphEngine]] = MemgraphEngine
    uri: str
    username: str
    password: str

    @model_validator(mode="before")
    @classmethod
    def populate_defaults(cls, data: Any) -> Any:
        load_dotenv()

        if data.get("uri") is None:
            data["uri"] = os.getenv("MEMGRAPH_URI")

        if data.get("username") is None:
            data["username"] = os.getenv("MEMGRAPH_USERNAME")

        if data.get("password") is None:
            data["password"] = os.getenv("MEMGRAPH_PASSWORD")

        return data



from __future__ import annotations

from datetime import date, datetime, time, timedelta
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional

from pydantic import BaseModel

from ..gql import gql_identifier_adapter, int_adapter
from ..result import NeontologyResult

if TYPE_CHECKING:
    from ..basenode import BaseNode
    from ..baserelationship import BaseRelationship


class GraphEngineBase:
    _supported_types: ClassVar[Any] = (
        list,
        bool,
        int,
        bytearray,
        float,
        str,
        bytes,
        date,
        time,
        datetime,
        timedelta,
    )

    def __init__(self, config: Optional[dict]) -> None:
        """Initialise connection to the engine

        Args:
            config (Optional[dict]): _description_
        """
        pass

    @classmethod
    def _export_type_converter(cls, value: Any) -> Any:
        if isinstance(value, dict):
            raise TypeError("Neontology doesn't support dict types for properties.")

        elif isinstance(value, (tuple, set)):
            new_value = list(value)
            return cls._export_type_converter(new_value)

        elif isinstance(value, list):
            if len(value) == 0:
                return []
            # items in a list must all be the same type
            item_type = type(value[0])
            for item in value:
                if isinstance(item, item_type) is False:
                    raise TypeError(
                        "For neo4j, all items in a list must be of the same type."
                    )

            return [cls._export_type_converter(x) for x in value]

        elif isinstance(value, cls._supported_types) is False:
            return str(value)

        else:
            return value

    @classmethod
    def export_dict_converter(cls, original_dict: Dict[str, Any]) -> Dict[str, Any]:
        """_summary_

        Args:
            export_dict (Dict[str, Any]): _description_

        Returns:
            Dict[str, Any]: _description_
        """

        export_dict = original_dict.copy()

        for k, v in export_dict.items():
            export_dict[k] = cls._export_type_converter(v)

        return export_dict

    def verify_connection(self) -> bool:
        raise NotImplementedError

    def close_connection(self) -> None:
        raise NotImplementedError

    def evaluate_query(
        self,
        cypher: str,
        params: Dict[str, Any] = {},
        node_classes: dict = {},
        relationship_classes: dict = {},
    ) -> NeontologyResult:
        raise NotImplementedError

    def evaluate_query_single(self, cypher: str, params: Dict[str, Any]) -> Any:
        raise NotImplementedError

    def apply_constraint(self, label: str, property: str) -> None:
        raise NotImplementedError

    def drop_constraint(self, constraint_name: str) -> None:
        raise NotImplementedError

    def get_constraints(self) -> list:
        raise NotImplementedError

    def create_nodes(
        self, labels: list, pp_key: str, properties: list, node_class: type["BaseNode"]
    ) -> List["BaseNode"]:
        """
        Args:
            labels (list): a list of labels to give created nodes
            pp_key (str): the primary property for the nodes
            properties (list): A list of dictionaries representing each node to be created.
                two keys with associated values pp (the value to assign the primary property)
                and props (dict with key value pairs for all other properties).

        Returns:
            list: list of created Nodes
        """

        label_identifiers = [gql_identifier_adapter.validate_strings(x) for x in labels]

        cypher = f"""
        UNWIND $node_list AS node
        create (n:{":".join(label_identifiers)} {{{gql_identifier_adapter.validate_strings(pp_key)}: node.pp}})
        SET n += node.props
        RETURN n
        """

        params = {"node_list": properties}

        node_classes = {node_class.__primarylabel__: node_class}

        results = self.evaluate_query(cypher, params, node_classes)

        return results.nodes

    def merge_nodes(
        self, labels: list, pp_key: str, properties: list, node_class: type["BaseNode"]
    ) -> List["BaseNode"]:
        """_summary_

        Args:
            labels (list): _description_
            pp_key (str): _description_
            properties (list): A list of dictionaries representing each node to be created.
                four keys with associated values: pp (the value to assign the primary property)
                set_on_match, set_on_create and always_set (dicts with key value pairs for all other properties).

        Returns:
            list: list of merged Nodes
        """

        label_identifiers = [gql_identifier_adapter.validate_strings(x) for x in labels]

        cypher = f"""
        UNWIND $node_list AS node
        MERGE (n:{":".join(label_identifiers)} {{{gql_identifier_adapter.validate_strings(pp_key)}: node.pp}})
        ON MATCH SET n += node.set_on_match
        ON CREATE SET n += node.set_on_create
        SET n += node.always_set
        RETURN n
        """

        params = {"node_list": properties}

        node_classes = {node_class.__primarylabel__: node_class}

        results = self.evaluate_query(cypher, params, node_classes)

        return results.nodes

    def delete_nodes(self, label: str, pp_key: str, pp_values: List[Any]) -> None:
        cypher = f"""
        UNWIND $pp_values AS pp
        MATCH (n:{gql_identifier_adapter.validate_strings(label)})
        WHERE n.{gql_identifier_adapter.validate_strings(pp_key)} = pp
        DETACH DELETE n
        """

        params = {"pp_values": pp_values}

        self.evaluate_query_single(cypher, params)

    def merge_relationships(
        self,
        source_label: str,
        target_label: str,
        source_prop: str,
        target_prop: str,
        rel_type: str,
        merge_on_props: List[str],
        rel_props: List[dict],
    ) -> None:
        # build a string of properties to merge on "prop_name: $prop_name"
        merge_props = ", ".join(
            [
                f"{gql_identifier_adapter.validate_strings(x)}: rel.{x}"
                for x in merge_on_props
            ]
        )

        cypher = f"""
        UNWIND $rel_list AS rel
        MATCH (source:{gql_identifier_adapter.validate_strings(source_label)})
        WHERE source.{gql_identifier_adapter.validate_strings(source_prop)} = rel.source_prop
        MATCH (target:{gql_identifier_adapter.validate_strings(target_label)})
        WHERE target.{gql_identifier_adapter.validate_strings(target_prop)} = rel.target_prop
        MERGE (source)-[r:{gql_identifier_adapter.validate_strings(rel_type)} {{ {merge_props} }}]->(target)
        ON MATCH SET r += rel.set_on_match
        ON CREATE SET r += rel.set_on_create
        SET r += rel.always_set
        """

        params = {"rel_list": rel_props}

        self.evaluate_query_single(cypher, params)

    def match_nodes(
        self,
        node_class: type["BaseNode"],
        limit: Optional[int] = None,
        skip: Optional[int] = None,
    ) -> List["BaseNode"]:
        """Get nodes of this type from the database.

        Run a MATCH cypher query to retrieve any Nodes with the label of this class.

        Args:
            limit (int, optional): Maximum number of results to return. Defaults to None.
            skip (int, optional): Skip through this many results (for pagination). Defaults to None.

        Returns:
            Optional[List[B]]: A list of node instances.
        """

        cypher = f"""
        MATCH(n:{node_class.__primarylabel__})
        RETURN n
        """

        params = {}

        if skip:
            cypher += " SKIP $skip "
            params["skip"] = int_adapter.validate_python(skip)

        if limit:
            cypher += " LIMIT $limit "
            params["limit"] = int_adapter.validate_python(limit)

        result = self.evaluate_query(
            cypher, params, node_classes={node_class.__primarylabel__: node_class}
        )

        return result.nodes

    def match_relationships(
        self,
        relationship_class: type["BaseRelationship"],
        limit: Optional[int] = None,
        skip: Optional[int] = None,
    ) -> List["BaseRelationship"]:
        """Get nodes of this type from the database.

        Run a MATCH cypher query to retrieve any Nodes with the label of this class.

        Args:
            limit (int, optional): Maximum number of results to return. Defaults to None.
            skip (int, optional): Skip through this many results (for pagination). Defaults to None.

        Returns:
            Optional[List[B]]: A list of node instances.
        """

        from ..utils import get_node_types, get_rels_by_type

        cypher = f"""
        MATCH (n)-[r:{relationship_class.__relationshiptype__}]->(o)
        RETURN DISTINCT n, r, o
        """

        params = {}

        if skip:
            cypher += " SKIP $skip "
            params["skip"] = int_adapter.validate_python(skip)

        if limit:
            cypher += " LIMIT $limit "
            params["limit"] = int_adapter.validate_python(limit)

        rel_types = get_rels_by_type()
        node_classes = get_node_types()

        result = self.evaluate_query(
            cypher,
            params,
            node_classes=node_classes,
            relationship_classes=rel_types,
        )

        return result.relationships


class GraphEngineConfig(BaseModel):
    engine: ClassVar[type[GraphEngineBase]]



import itertools
import os
import warnings
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional

from dotenv import load_dotenv
from neo4j import GraphDatabase
from neo4j import Record as Neo4jRecord
from neo4j import Result as Neo4jResult
from neo4j.graph import Node as Neo4jNode
from neo4j.graph import Path as Neo4jPath
from neo4j.graph import Relationship as Neo4jRelationship
from neo4j.time import Date as Neo4jDate
from neo4j.time import DateTime as Neo4jDateTime
from neo4j.time import Time as Neo4jTime
from pydantic import model_validator

from ..gql import gql_identifier_adapter
from ..result import NeontologyResult
from .graphengine import GraphEngineBase, GraphEngineConfig

if TYPE_CHECKING:
    from ..basenode import BaseNode
    from ..baserelationship import BaseRelationship, RelationshipTypeData


def convert_neo4j_types(input_dict: dict) -> dict:
    output_dict = dict(input_dict)

    for key in output_dict:
        try:
            new_date = output_dict[key].to_native()

            if isinstance(output_dict[key], (Neo4jDateTime, Neo4jDate, Neo4jTime)):
                output_dict[key] = new_date

        except AttributeError:
            pass

    return output_dict


def neo4j_node_to_neontology_node(
    neo4j_node: Neo4jNode, node_classes: dict
) -> Optional["BaseNode"]:
    node_labels = list(neo4j_node.labels)

    primary_labels = set(node_labels).intersection(set(node_classes.keys()))

    secondary_labels = set(node_labels).difference(set(node_classes.keys()))

    if len(primary_labels) == 1:
        primary_label = primary_labels.pop()

        node_dict = convert_neo4j_types(dict(neo4j_node))

        node = node_classes[primary_label](**node_dict)

        # warn if the secondary labels aren't what's expected

        if set(node.__secondarylabels__) != secondary_labels:
            warnings.warn(f"Unexpected secondary labels returned: {secondary_labels}")

        return node

    # gracefully handle cases where we don't have a class defined
    # for the identified label or where we get more than one valid primary label
    else:
        warnings.warn(f"Unexpected primary labels returned: {primary_labels}")

        return None


def neo4j_relationship_to_neontology_rel(
    neo4j_rel: Neo4jRelationship, node_classes: dict, rel_classes: dict
) -> Optional["BaseRelationship"]:
    rel_type = neo4j_rel.type
    rel_type_data = rel_classes[rel_type]

    if not rel_type_data:
        warnings.warn(
            (
                f"Could not find a class for {rel_type} relationship type."
                " Did you define the class before initializing Neontology?"
            )
        )
        return None

    if (
        not neo4j_rel.start_node
        or not neo4j_rel.start_node.labels
        or not neo4j_rel.end_node
        or not neo4j_rel.end_node.labels
    ):
        warnings.warn(
            (
                f"{rel_type} relationship type query did not include nodes."
                " To get neontology relationships, return source and target "
                "nodes as part of result."
            )
        )
        return None

    src_node = neo4j_node_to_neontology_node(neo4j_rel.start_node, node_classes)
    tgt_node = neo4j_node_to_neontology_node(neo4j_rel.end_node, node_classes)

    rel_props = convert_neo4j_types(dict(neo4j_rel))
    rel_props["source"] = src_node
    rel_props["target"] = tgt_node

    return rel_type_data.relationship_class(**rel_props)


def neo4j_records_to_neontology_records(
    records: List[Neo4jRecord],
    node_classes: dict,
    rel_classes: Dict[str, "RelationshipTypeData"],
) -> tuple:
    new_records = []

    for record in records:
        new_record: Dict[str, dict] = {"nodes": {}, "relationships": {}, "paths": {}}

        for key, entry in record.items():
            if isinstance(entry, Neo4jNode):
                neontology_node = neo4j_node_to_neontology_node(entry, node_classes)

                if neontology_node:
                    new_record["nodes"][key] = neontology_node

            elif isinstance(entry, Neo4jRelationship):
                neontology_rel = neo4j_relationship_to_neontology_rel(
                    entry, node_classes, rel_classes
                )

                if neontology_rel:
                    new_record["relationships"][key] = neontology_rel

            elif isinstance(entry, Neo4jPath):
                entry_path = []
                for step in entry:
                    step_rel = neo4j_relationship_to_neontology_rel(
                        step, node_classes, rel_classes
                    )
                    entry_path.append(step_rel)
                if entry_path:
                    new_record["paths"][key] = entry_path

        new_records.append(new_record)

    nodes_list_of_lists = [x["nodes"].values() for x in new_records]

    nodes = list(itertools.chain.from_iterable(nodes_list_of_lists))

    nodes_map = {f"{x.__primarylabel__}:{x.get_pp()}": x for x in nodes}

    unique_nodes = list(nodes_map.values())

    rels_list_of_lists = [x["relationships"].values() for x in new_records]
    rels = list(itertools.chain.from_iterable(rels_list_of_lists))

    paths_list_of_lists = [x["paths"].values() for x in new_records]
    paths = list(itertools.chain.from_iterable(paths_list_of_lists))

    return new_records, unique_nodes, rels, paths


class Neo4jEngine(GraphEngineBase):
    def __init__(self, config: "Neo4jConfig") -> None:
        """Initialise connection to the engine

        Args:
            config - Takes a Neo4jConfig object.
        """

        self.driver = GraphDatabase.driver(  # type: ignore
            config.uri,
            auth=(config.username, config.password),
        )

    def verify_connection(self) -> bool:
        try:
            self.driver.verify_connectivity()
            return True
        except:  # noqa: E722
            return False

    def close_connection(self) -> None:
        try:
            self.driver.close()
        except AttributeError:
            pass

    def evaluate_query(
        self,
        cypher: str,
        params: dict = {},
        node_classes: dict = {},
        relationship_classes: dict = {},
    ) -> NeontologyResult:
        result = self.driver.execute_query(cypher, parameters_=params)

        neo4j_records = result.records
        neontology_records, nodes, rels, paths = neo4j_records_to_neontology_records(
            neo4j_records, node_classes, relationship_classes
        )

        return NeontologyResult(
            records_raw=neo4j_records,
            records=neontology_records,
            nodes=nodes,
            relationships=rels,
            paths=paths,
        )

    def evaluate_query_single(self, cypher: str, params: dict = {}) -> Optional[Any]:
        result = self.driver.execute_query(
            cypher, parameters_=params, result_transformer_=Neo4jResult.single
        )

        if result:
            return result.value()

        else:
            return None

    def apply_constraint(self, label: str, property: str) -> None:
        cypher = f"""
        CREATE CONSTRAINT IF NOT EXISTS
        FOR (n:{gql_identifier_adapter.validate_strings(label)})
        REQUIRE n.{gql_identifier_adapter.validate_strings(property)} IS UNIQUE
        """

        self.evaluate_query_single(cypher)

    def drop_constraint(self, constraint_name: str) -> None:
        drop_cypher = f"""
        DROP CONSTRAINT {gql_identifier_adapter.validate_strings(constraint_name)}
        """
        self.evaluate_query_single(drop_cypher)

    def get_constraints(self) -> list:
        get_constraints_query = """
        SHOW CONSTRAINTS yield name
        RETURN COLLECT(DISTINCT name)
        """

        constraints = self.evaluate_query_single(get_constraints_query)

        if constraints:
            return constraints

        else:
            return []


class Neo4jConfig(GraphEngineConfig):
    engine: ClassVar[type[Neo4jEngine]] = Neo4jEngine
    uri: str
    username: str
    password: str

    @model_validator(mode="before")
    @classmethod
    def populate_defaults(cls, data: Any) -> Any:
        load_dotenv()

        if data.get("uri") is None:
            data["uri"] = os.getenv("NEO4J_URI")

        if data.get("username") is None:
            data["username"] = os.getenv("NEO4J_USERNAME")

        if data.get("password") is None:
            data["password"] = os.getenv("NEO4J_PASSWORD")

        return data



import itertools
import json
import warnings
from typing import Any, ClassVar, Dict, List, Optional, Type, TypeVar

import numpy as np
import pandas as pd
from pydantic import BaseModel, PrivateAttr, ValidationError, model_validator

from neontology.graphconnection import GraphConnection

from .basenode import BaseNode
from .commonmodel import CommonModel
from .gql import gql_identifier_adapter
from .schema_utils import RelationshipSchema, SchemaProperty, extract_type_mapping

R = TypeVar("R", bound="BaseRelationship")


class BaseRelationship(CommonModel):  # pyre-ignore[13]
    source: BaseNode
    target: BaseNode

    __relationshiptype__: ClassVar[Optional[str]] = None

    _merge_on: List[str] = (
        PrivateAttr()
    )  # what relationship properties should we merge on

    def __init__(self, **data: dict):
        super().__init__(**data)

        # we can define 'abstract' relationships which don't have a label
        # these are to provide common properties to be used by subclassed relationships
        # but shouldn't be put in the graph or even instantiated
        if self.__relationshiptype__ is None:
            raise NotImplementedError(
                "Relationships to be used in the graph must define a relationship type."
            )

    @classmethod
    def _set_prop_usage(cls) -> None:
        super()._set_prop_usage()
        cls._merge_on = cls._get_prop_usage("merge_on")

    @model_validator(mode="after")
    def validate_identifiers(self) -> "BaseRelationship":
        try:
            gql_identifier_adapter.validate_strings(self.__relationshiptype__)
        except AttributeError:
            pass
        except ValidationError:
            warnings.warn(
                (
                    "Relationship Type should contain only alphanumeric characters and underscores."
                    " It should begin with an alphabetic character."
                )
            )

        return self

    @classmethod
    def get_relationship_type(cls) -> Optional[str]:
        """Get the relationship type to use for creating and matching this relationship.

        If __relationship__ has been specified, use that.

        Otherwise use the class name in uppercase

        Returns:
            str: the string to use for creating and matching this relationship
        """
        return cls.__relationshiptype__  # pyre-ignore[7]

    def _get_merge_parameters(
        self, source_prop: str, target_prop: str
    ) -> Dict[str, Any]:
        """

        Returns:
            Dict[str, Any]: a dictionary of key/value pairs.
        """

        exclusions = {"source", "target"}

        # get all the properties
        all_props = self.model_dump(exclude=exclusions)

        # merge_props properties will be referenced individually with kwargs
        merge_props = {k: all_props[k] for k in self._merge_on}

        always_set = {k: all_props[k] for k in self._always_set}

        set_on_match = {k: all_props[k] for k in self._set_on_match}

        set_on_create = {k: all_props[k] for k in self._set_on_create}

        source_prop = self.source.model_dump()[source_prop]
        target_prop = self.target.model_dump()[target_prop]

        params = {
            "source_prop": source_prop,
            "target_prop": target_prop,
            "always_set": always_set,
            "set_on_match": set_on_match,
            "set_on_create": set_on_create,
            **merge_props,
        }

        return params

    def merge(
        self,
    ) -> None:
        """Merge this relationship into the database."""
        source_label = self.source.__primarylabel__
        target_label = self.target.__primarylabel__

        if not source_label or not target_label:
            raise ValueError(
                "Source and target Nodes must have a defined primary label for creating a relationship."
            )

        source_prop = self.source.__primaryproperty__
        target_prop = self.target.__primaryproperty__

        rel_props = self._get_merge_parameters(
            source_prop=source_prop, target_prop=target_prop
        )

        merge_on_props = self._merge_on

        rel_type = self.get_relationship_type()

        if not rel_type:
            raise ValueError(
                "Realtionship must have a defined relationship type for creating a relationship."
            )

        gc = GraphConnection()

        gc.merge_relationships(
            source_label,
            target_label,
            source_prop,
            target_prop,
            rel_type,
            merge_on_props,
            [rel_props],
        )

    @classmethod
    def merge_relationships(
        cls: Type[R],
        rels: List[R],
        source_prop: Optional[str] = None,
        target_prop: Optional[str] = None,
    ) -> None:
        """Merge multiple relationships (of this type) into the database.

        Sometimes the source and target label may be ambiguous (e.g. where we have subclassed nodes)
            In this case you can explicitly pass in the relevant types

        Sometimes we want to match nodes on a property which isn't the primary property,
        so we can specify what property to use.

        Args:
            cls (Type[R]): this class
            rels (List[R]): a list of relationships which are instances of this class

        Raises:
            TypeError: If relationships are provided which aren't of this class
        """

        # define the properties to merge on
        merge_on_props = cls._get_prop_usage("merge_on")

        # sources and targets could have different primary labels
        # to operate efficiently, we group like source and targets for batch creation of relationships
        grouped_rels = itertools.groupby(
            rels, lambda x: (x.source.__class__, x.target.__class__)
        )

        for node_clases, common_rels in grouped_rels:
            src_class = node_clases[0]
            tgt_class = node_clases[1]

            source_label = src_class.__primarylabel__
            target_label = tgt_class.__primarylabel__

            if source_prop is None:
                source_prop = src_class.__primaryproperty__

            if target_prop is None:
                target_prop = tgt_class.__primaryproperty__

            if not source_label or not target_label:
                raise ValueError(
                    "Source and target Nodes must have a defined primary label to create a relationship."
                )

            rel_type = cls.get_relationship_type()

            if not rel_type:
                raise ValueError(
                    "Relationship must have a defined relationship type for creating a relationship."
                )

            rel_list: List[Dict[str, Any]] = [
                x._get_merge_parameters(source_prop, target_prop) for x in common_rels
            ]

            gc = GraphConnection()

            gc.merge_relationships(
                source_label,
                target_label,
                source_prop,
                target_prop,
                rel_type,
                merge_on_props,
                rel_list,
            )

    @classmethod
    def merge_records(
        cls: Type[R],
        records: List[Dict[str, Any]],
        source_type: Optional[Type[BaseNode]] = None,
        target_type: Optional[Type[BaseNode]] = None,
        source_prop: Optional[str] = None,
        target_prop: Optional[str] = None,
    ) -> None:
        """Take a list of dictionaries and use them to merge in relationships in the graph.

        Sometimes, a relationship can accept nodes which subclass a particular node type.
            In these instances, it may be necessary to explicitly state what type of node should be used.

        Each record should have a source and target key where the value is the primary property
            value of the respective nodes.

        Args:
            records (List[Dict[str, Any]]): a list of dictionaries used to populate relationships
            source_type: explicitly state the class to use for source node
            target_type: explicitly state the class to use for target node
        """

        hydrated_list = []

        if source_type is None:
            source_type = cls.model_fields["source"].annotation

        if target_type is None:
            target_type = cls.model_fields["target"].annotation

        if not source_type or not target_type:
            raise ValueError("Source and target Nodes types not defined.")

        if source_prop is None:
            source_prop = source_type.__primaryproperty__

        if target_prop is None:
            target_prop = target_type.__primaryproperty__

        for record in records:
            hydrated = dict(record)

            hydrated["source"] = source_type.model_construct(
                **{source_prop: record["source"]}
            )
            hydrated["target"] = target_type.model_construct(
                **{target_prop: record["target"]}
            )

            hydrated_list.append(cls(**hydrated))

        cls.merge_relationships(
            hydrated_list,
            source_prop=source_prop,
            target_prop=target_prop,
        )

    @classmethod
    def merge_df(
        cls: Type[R],
        df: pd.DataFrame,
        source_type: Optional[Type[BaseNode]] = None,
        target_type: Optional[Type[BaseNode]] = None,
        source_prop: Optional[str] = None,
        target_prop: Optional[str] = None,
    ) -> None:
        """Merge in relationships based on data in a pandas data frame

        Expects columns named 'source' and 'target' with the primary property value
            for the source and target nodes.

        Then additional fields should have a corresponding column.

        Args:
            df (pd.DataFrame): pandas dataframe where each row represents a relationship to merge
        """

        if df.empty is False:
            records = df.replace([np.nan], None).to_dict(orient="records")
            cls.merge_records(
                records,
                source_type=source_type,
                source_prop=source_prop,
                target_type=target_type,
                target_prop=target_prop,
            )

    @classmethod
    def match_relationships(
        cls, limit: Optional[int] = None, skip: Optional[int] = None
    ) -> List["BaseRelationship"]:
        gc = GraphConnection()
        result = gc.match_relationships(cls, limit, skip)

        return result

    @classmethod
    def get_count(cls):
        gc = GraphConnection()
        cypher = (
            f"MATCH (n)-[r:{cls.__relationshiptype__}]->(o) RETURN COUNT(DISTINCT r)"
        )
        result = gc.evaluate_query_single(cypher)
        return result

    def _prep_dump_dict(self, dumped_model: dict, exclude_node_props: True) -> dict:
        if exclude_node_props is True:
            dumped_model["source"] = self.source.get_pp()
            dumped_model["SOURCE_LABEL"] = self.source.__primarylabel__
            dumped_model["target"] = self.target.get_pp()
            dumped_model["TARGET_LABEL"] = self.target.__primarylabel__

        else:
            dumped_model["source"]["LABEL"] = self.source.__primarylabel__
            dumped_model["source"]["PK"] = self.source.get_pp()

            dumped_model["target"]["LABEL"] = self.target.__primarylabel__
            dumped_model["target"]["PK"] = self.target.get_pp()

        dumped_model["RELATIONSHIP_TYPE"] = self.__relationshiptype__

        return dumped_model

    def neontology_dump(
        self,
        exclude_node_props: bool = True,
        exclude: Optional[set] = None,
        exclude_none: bool = True,
        **kwargs,
    ) -> dict:
        dumped_model = self.model_dump(
            exclude_none=exclude_none, exclude=exclude, **kwargs
        )

        return self._prep_dump_dict(dumped_model, exclude_node_props)

    def neontology_dump_json(
        self,
        exclude_node_props: bool = True,
        exclude: Optional[set] = None,
        exclude_none: bool = True,
        **kwargs,
    ) -> str:
        # pydantic converts values to be json serializable, make use of this first
        original_json = self.model_dump_json(
            exclude_none=exclude_none, exclude=exclude, **kwargs
        )
        model_dict = json.loads(original_json)

        return json.dumps(self._prep_dump_dict(model_dict, exclude_node_props))

    @classmethod
    def neontology_schema(
        cls,
        source_labels: Optional[List[str]] = None,
        target_labels: Optional[List[str]] = None,
    ) -> RelationshipSchema:
        schema_properties = []
        rel_type = cls.__relationshiptype__

        if not rel_type:
            raise ValueError("Relationship doesn't have a relationship type.")

        for field_name, field_props in cls.model_fields.items():
            if field_name in ["source", "target"]:
                continue

            field_type = extract_type_mapping(
                field_props.annotation, show_optional=True
            )

            required_field = field_props.is_required()

            rel_prop = SchemaProperty(
                type_annotation=field_type,
                name=field_name,
                required=required_field,
            )

            if required_field is True:
                schema_properties.insert(0, rel_prop)

            else:
                schema_properties.append(rel_prop)

        source_type = cls.model_fields["source"].annotation
        target_type = cls.model_fields["target"].annotation

        if not target_labels:
            if (
                hasattr(target_type, "__primarylabel__")
                and target_type.__primarylabel__
            ):
                target_labels = [target_type.__primarylabel__]

        if not source_labels:
            if (
                hasattr(source_type, "__primarylabel__")
                and source_type.__primarylabel__
            ):
                source_labels = [source_type.__primarylabel__]

        if not source_labels or not target_labels:
            from neontology.utils import get_node_types

            if not target_labels:
                # return concrete subclasses of the abstract node class given
                retrieved_node_types = get_node_types(target_type)
                target_labels = list(retrieved_node_types.keys())

            if not source_labels:
                # return concrete subclasses of the abstract node class given
                retrieved_node_types = get_node_types(source_type)
                source_labels = list(retrieved_node_types.keys())

        schema = RelationshipSchema(
            name=rel_type,
            relationship_type=rel_type,
            properties=schema_properties,
            target_labels=target_labels,
            source_labels=source_labels,
        )

        return schema


class RelationshipTypeData(BaseModel):
    relationship_class: type[BaseRelationship]
    source_class: type[BaseNode]
    target_class: type[BaseNode]
    all_source_classes: List[type[BaseNode]]
    all_target_classes: List[type[BaseNode]]
